from django.db import models
from django.utils.encoding import python_2_unicode_compatible

__all__ = ['Purchase']


@python_2_unicode_compatible
class Purchase(models.Model):
    book = models.ForeignKey('Book')
    buyer_ip = models.GenericIPAddressField()
    purchased_at = models.DateTimeField()

    def __str__(self):
        return "{0} bought {1} at {2}".format(self.buyer_ip, self.book, self.purchased_at)

    class Meta:
        sensitive_fields = {'buyer_ip', 'purchased_at'}



try:
    from django.apps import apps
except ImportError:
    from django.db import models as apps
from django.core.management.base import BaseCommand, CommandError
from scrubpii.utils import get_updates_for_model


class Command(BaseCommand):
    can_import_settings = True
    output_transaction = True

    def handle(self, *args, **options):
        models = apps.get_models()
        script = ""
        for model in models:
            updates = get_updates_for_model(model)
            if updates:
                script += updates
        return script



from os.path import join as pjoin, abspath, dirname, pardir


SECRET_KEY = 'SECRET'
PROJ_ROOT = abspath(pjoin(dirname(__file__), pardir))

ADMINS = (
    ('Example', 'tests@example.net'),
)
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}
MEDIA_ROOT = pjoin(PROJ_ROOT, 'media')
MEDIA_URL = '/media/'
ROOT_URLCONF = 'tests.urls'
INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'tests.testapp',
    'scrubpii',
)

TEMPLATE_CONTEXT_PROCESSORS = (
    'django.core.context_processors.request',
)
MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)



import uuid
from django.db import connection
from django.conf import settings
from django.contrib.auth.hashers import make_password

salt = None

UPDATE_QUERY_TEMPLATE = """
    UPDATE {table_name}
    SET {assignments};
"""
ASSIGNMENT_TEMPLATE = """
        {field_name} = {value_method}"""

OBFUSCATED_SOURCES = {
    "CharField": {
        "postgresql": "substring(md5(random()::text) FROM 0 FOR 10)",
        "default":    "'###########'",
    },
    "TextField": {
        "default":    ("'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Fusce urna lorem, condimentum ut"
                       " pretium commodo, iaculis eget lacus. In eget lorem varius erat auctor tempor. Pellentesque"
                       " lacinia varius erat, sed eleifend magna consectetur et. Maecenas porttitor varius enim quis"
                       " tincidunt. Maecenas feugiat ipsum a augue iaculis, a volutpat urna egestas. Cras molestie"
                       " mauris non arcu tempor efficitur. Donec at felis lacus. Nulla posuere pretium magna at"
                       " elementum.'"),
    },
    "EmailField": {
        "sqlite":     "'user-' || id || '@example.net'",
        "default":    "concat('user-', id, '@example.net')",
    },
    "DateField": {
        "postgresql":  "date '1990-06-02'",
        "sqlite":      "date('1990-06-02')",
    },
    "DateTimeField": {
        "default":    "timestamp '1990-06-02 10:00:00'",
        "sqlite":      "datetime('1990-06-02 10:00:00')",
    },
    "GenericIPAddressField": {
        "postgresql": """CONCAT(
                             '2001:db8:',
                             SUBSTRING(md5(concat({field_name}, '{salt}')) FROM  1 FOR 4), ':',
                             SUBSTRING(md5(concat({field_name}, '{salt}')) FROM  5 FOR 4), ':',
                             SUBSTRING(md5(concat({field_name}, '{salt}')) FROM  9 FOR 4), ':',
                             SUBSTRING(md5(concat({field_name}, '{salt}')) FROM 13 FOR 4), ':',
                             SUBSTRING(md5(concat({field_name}, '{salt}')) FROM 17 FOR 4), ':',
                             SUBSTRING(md5(concat({field_name}, '{salt}')) FROM 21 FOR 4)
                         )::inet""",
        "default":    "'2001:db8::1'"
    },
    "username": {
        "sqlite":     "'user-' || id",
        "default":    "concat('user-', id)",
    },
    "password": {
        "default":    "'{0}'".format(make_password('password', hasher='md5')),
    }
}


def get_value_method(field, database='default'):
    try:
        options = OBFUSCATED_SOURCES[field.column]
    except KeyError:
        options = OBFUSCATED_SOURCES[type(field).__name__]
    if database in options:
        return options[database]
    return options['default']


def get_sensitive_fields(klass):
    annotated = getattr(klass._meta, 'sensitive_fields', set())
    additional_fields = getattr(settings, 'SCRUB_PII_ADDITIONAL_FIELDS', None)
    if additional_fields is not None:
        class_ref = "{0}.{1}".format(klass._meta.app_label, klass.__name__)
        annotated = annotated.union(additional_fields.get(class_ref, set()))
    return annotated


def get_updates_for_model(klass):
    global salt
    if salt is None:
        salt = uuid.uuid4()

    fields = []
    sensitive_fields = get_sensitive_fields(klass)
    if not sensitive_fields:
        return None
    for field in sensitive_fields:
        field_object = klass._meta.get_field(field)
        value_method = get_value_method(field_object, connection.vendor)
        data = {'table_name': klass._meta.db_table, 'field_name': field, 'salt': salt}
        fields.append({'field_name': field, 'value_method': value_method.format(**data)})
    assignments = map(lambda x: ASSIGNMENT_TEMPLATE.format(**x), fields)
    assignments = ",".join(assignments)
    query = UPDATE_QUERY_TEMPLATE.format(table_name=klass._meta.db_table, assignments=assignments)
    return query



from .utils import get_sensitive_fields  # noqa

import django.db.models.options as options


class allow_sensitive_fields(object):
    """Patches the Django Options class to allow for setting sensitive_fields."""

    def __enter__(self):
        self.original_names = options.DEFAULT_NAMES
        if 'sensitive_fields' in options.DEFAULT_NAMES:
            raise ValueError("sensitive_fields is already defined in Django!")
        options.DEFAULT_NAMES = options.DEFAULT_NAMES + ('sensitive_fields',)

    def __exit__(self, type, value, traceback):
        options.DEFAULT_NAMES = self.original_names
